.TH Customizing TCL "" "TCL - Tool Command Language"
.ad b
.SH OVERVIEW
This document is intended to help the programmer who wishes to extend
Tcl with C language routines.  It should also be useful to someone
wishing to add Tcl to an existing editor, comm program, etc.  There
is also programming information in the \fITcl.man\fR manual in the
\fIbaseline/man\fR
directory.
.SH WRITING TCL EXTENSIONS IN C
C extensions to Tcl must be written to receive their arguments in the
manner Tcl uses to pass them.
.PP
A C routine is called from Tcl with four arguments, a client data pointer,
an interpreter pointer, an argument count and a pointer to an array of
pointers to character strings containing the Tcl arguments to the routine.
.PP
A Tcl extension in C is now presented, and described below:
.sp
.nf
.ft CW
#include "tcl.h"

int Tcl_EchoCmd(clientData, interp, argc, argv)
void *clientData;
Tcl_Interp *interp;
int argc;
char **argv;
{
	int i;

	for (i = 1; i < argc; i++) {
		printf("%s ",argv[i]);
	}
	printf("\n");
	return TCL_OK;
}
.ft R
.fi
.PP      

.PP
The client data pointer will be described later.
.PP
The interpreter pointer is the "key" to an interpreter.  It is returned
by \fITcl_CreateInterpreter\fR and is passed around within Tcl and by your
C code.  The structure pointed to by the interpreter pointer, and all
of the subordinate structures that branch off of it, make up an interpreter
context, which includes all of the currently defined procedures, commands,
variables, arrays and the execution state of that interpreter.
.PP
The argument count and pointer to an array of pointers to textual arguments
is handled by your C code in the same manner that you would use in writing a
C \fImain\fR function -- the argument count and array of pointers
works the same
as in a C \fImain\fR call; pointers to the arguments to the function are
contained
in the \fIargv\fR array.  Similar to a C main, the first argument
(\fIargv[0]\fR) is the
name the routine was called as (in a main, the name the program
was invoked as).
.PP
In the above example, all of the arguments are output with a space between
each one by looping through \fIargv\fR from one to the argument count,
\fIargc\fR, and a
newline terminates the line -- an echo command, although a "real" echo
command would not add a trailing blank like this one does.
.PP
All arguments from a Tcl call to a Tcl C extension are passed as strings.
If your C routine expects certain numeric arguments, your routine must first
convert them using the \fIStrToInt\fR or 
\fIStrToLong\fR functions, or some other method
of your own devising.  If you program produces a numeric result, it should
return a string equivalent to that numeric value.  A popular way of doing
this is something like...
.sp
.nf
.ft CW
	sprintf(interp->result, "%ld", result);
.ft R
.fi
.PP      
More sophisticated commands should verify their arguments when possible,
both by examining the argument count, by verifying that numeric fields
are really numeric, that values
are in range when their ranges are known, and so forth.
.PP
We have been careful to try to make Tcl as bullet-proof as possible, in the
sense that Tcl programs should not be able to cause Tcl to dump core.  Please
do the same with your C extensions by validating arguments as above.
.PP
In the command below, two or more arguments are compared and the one with
the maximum value is returned, if all goes well.  It is an error if there
are fewer than three arguments (the pointer to the "max" command text itself,
\fIargv[0]\fR, and pointers to at least two arguments to compare the
values of).
.PP
This routine also shows the use of the programmer labor-saving
\fITcl_RetError\fR
routine.  See the manual page on Tcl_RetError for details.
.sp
.nf
.ft CW
int
CmdMax (clientData, interp, argc, argv)
    char       *clientData;
    Tcl_Interp *interp;
    int         argc;
    char      **argv;
{
int MaxVal = MININT;
int MaxIdx = 1;
int Value, idx;
        

if (argc < 3)
        return Tcl_RetError (interp, "TCL", 
                             "bad # arg: %.50s num1 num2 [..numN]", argv[0]);

for (idx = 1; idx < argc; idx++)
        {
        if (!StrToInt (argv[idx], 10, &Value))
                return Tcl_RetError (interp, "TCL", 
                                     "invalid numeric argument");
        if (Value > MaxVal)
                {
                MaxVal = Value;
                MaxIdx = idx;
                }
        }
strcpy (interp->result, argv[MaxIdx]);
return TCL_OK;
}
.ft R
.fi
.PP      
When Tcl-callable functions complete, they should normally return
\fBTCL_OK\fR or \fBTCL_ERROR\fR.
\fBTCL_OK\fR is returned when the command succeeded and \fBTCL_ERROR\fR
is returned when the command has failed rather drastically.  
\fBTCL_ERROR\fR should
be returned for all syntax errors, non-numeric values where numeric ones
were expected, and so forth.  Less clear in some cases is whether Tcl errors
should be returned or whether a function should just return a status values.
For example, we have made end-of-file during a \fIgets\fR return a status,
but \fIopen\fR
returns an error if the open fails.  Errors can be caught
from Tcl programs using the Tcl \fIcatch\fB command.
.PP
Less common possible return values are 
\fITCL_RETURN\fR, \fITCL_BREAK\fR and \fITCL_CONTINUE\fR.
These would be useful to you if you were adding new control and/or looping
structures to Tcl.  To see these values in action, examine the source to
the \fIwhile\fR, \fIfor\fR, \fIif\fR and \fIloop\fR commands.
.PP
Note the use of \fITcl_RetError\fR to return an error above.  
\fITcl_RetError\fR
accepts
a printf-style variable number of arguments for generating the error text.
Study the manual page for \fITcl_RetError\fR to understand how to use this
programmer-time-saving method of coding error returns.
.PP
.SH INSTALLING YOUR COMMAND
To install your command into Tcl your must call 
\fITcl_CreateCommand\fR, passing
it the pointer into the interpreter you want to install the command into,
the name of the command, a pointer to the C function, a client data pointer,
and a pointer to an optional callback routine.
.PP
The client data pointer and the callback routine will be described later.
.PP
It is our convention in coding Tcl extension to create an "init" routine
that installs all of the Tcl-callable C routines for a given package into
a specified interpreter.
.PP
For example, for the Max function above (which incidentally comes from
math.c in the extend/src directory) is installed by the 
\fITcl_InitMath\fR
routine, which also comes from init.c...
.sp
.nf
.ft CW
int
/*
 *----------------------------------------------------------------------
 *
 *  Tcl_InitMath --
 *      Initialize the Tcl math extensions...
 *
 *----------------------------------------------------------------------
 */
Tcl_InitMath (interp)
    Tcl_Interp *interp;
{
Tcl_CreateCommand (interp, "max", CmdMax, (ClientData)NULL, 
                  (void (*)())NULL);
Tcl_CreateCommand (interp, "min", CmdMin, (ClientData)NULL, 
                  (void (*)())NULL);
Tcl_CreateCommand (interp, "random", CmdRandom, (ClientData)NULL, 
                  (void (*)())NULL);
Tcl_CreateCommand (interp, "seed", CmdSeed, (ClientData)NULL,
                  (void (*)())NULL);
}
.ft R
.fi
.PP
In the above example, the max, min, random and seed functions are added
to the specified interpreter.  The client data pointer and callback
function pointer are null.
.PP
The client data pointer provides a means for Tcl commands to have data
associated through them that is not global to the C program including
the Tcl core.  It is essential in a multi-interpreter environment
(one program has created and is making use of multiple Tcl interpreters)
for the C routines to maintain any permanent data they need relative to
each interpreter being used, or there would be reentrancy problems.
Tcl solves this through the client data mechanism.  When you are about
to call 
\fITcl_CreateCommand\fR to add a new command to an interpreter, if that
command needs to keep some read/write data from one invocation to another,
you should allocate the space, preferably using \fIckalloc\fR, then pass
the address of that space as the ClientData pointer to
\fITcl_CreateCommand\fR.
.PP
When your command is called from Tcl, the ClientData pointer you gave to 
\fITcl_CreateCommand\fR when you added the command to that interpreter
is passed
to your C routine through the ClientData pointer calling argument.
.PP
Commands that need to share this data with one another can do so by using the
same clientdata pointer when the commands are added.
.PP
It is very important to note that the Tcl extensions in the extend/src
directory have *not* had all of their data set up in this way, so at
the time of this writing (release 4.0) the Tcl extensions do not support
multiple interpreters within one invocation of Tcl.  The modification of
these routines to allow support for multiple interpreters is planned at
some point in the future.
.SH INTEL '286 GOTCHAS
The '286 programmer who is not using an ANSI C standard compiler with
function prototypes must be vigilant to ensure that anytime NULL is
passed to a function as a pointer it is explicitly cast to
\fB(void *)\fR or equivalent.  Also remember that Tcl math within
expressions is carried out to 32-bit, so that you should usually use
the \fIlong\fR variable type for your integers, \fIStrToLong\fR (rather
than \fIStrToInt\fR) to convert strings to long integers, and remember
to use \fB%ld\fR when printing results with \fIsprintf\fR and so
forth.
.PP
To maintain '286 compatibility, all C programmers are asked to follow
these guidelines.  I know you don't want to, but there are a lot of
286 machines out there and it is nice that they are able to run Tcl.
.SH THEORY OF HANDLES
Sometimes you need to have a data element that isn't readily representable
as a string within Tcl, for example a pointer to a complex C data structure.
We do not think it is a good idea to try to throw pointers around within
Tcl as strings by converting them to and from hex or integer representations,
for example.  It is too easy to screw one up and the likely outcome of
doing that is a core dump.
.PP
Instead what we have done is created the concept of \fIhandles\fR.
Handles are identifiers a C extension can pass back to, and accept from,
Tcl to make the transition between what your C code knows something as
and what name Tcl knows for it as safe and painless as possible.  For example,
the stdio package included in Tcl uses file handles.  When you open a
file from Tcl, a handle is returned of the form \fBfile\fIn\fR where
\fIn\fR is a file number.  When you pass the file handle back to \fIputs\fR,
\fIgets\fR, \fIseek\fR, \fIflush\fR and so forth, they validate the file
handle by checking the the file text is present, then convert the file
number to an integer that they use to look into a data structure of pointers
to Tcl open file structures, which contain a Unix file descriptor, flags
indicating whether or not the file is currently open, whether the file is
a file or a pipe and so forth.
.PP
The \fIconvfile\fR C subroutine in the Tcl stdio package is used to
perform the validation and conversion of a filehandle to an
index into the array of pointers to Tcl open file structures.
.PP
If you need a similar capability, it would be best to do something similar
to what we did in the stdio package.  We recommend that you use a
unique-to-your-package textual handle coupled with a specific identifier
and validate it thoroughly when it's passed back.  It is much easier to
track down a bug with an implicated handle named something like \fBfile4\fR
or \fBbitmap6\fR than just \fB6\fR.
.PP
.SH TRACKING MEMORY CORRUPTION PROBLEMS
Occasionally you may write code that scribbles past the end of an
allocated piece of memory and so forth.  The memory debugging
routines included in Tcl can help find these problems.  If you
think you have isolated the problem down to some range of lines,
it might help for you to stick in calls to \fIValidateAllMemory\fR.
It takes two parameters, a character string of the file name and
a character string to the line number -- they can really be anything
you want.
