/* 
 * Copyright 1990 Karl Lehenbauer, Mark Diekhans, 
 *                Peter da Silva and Jordan Henderson.
 *
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that the above copyright
 * notice appear in all copies.  Karl Lehenbauer, Mark Diekhans,
 * Peter da Silva and Jordan Henderson make no representations
 * about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 */

#ifndef NOSCCSID
        static char sccs_id[] = "@(#) %M% %I% %H%";
#endif

#include <signal.h>

#include <stdio.h>
#if defined(SYSV_3_2_2) || defined(XENIX386)
#  include <bool.h>
#endif

#include "tcl.h"

#ifdef MEM_DEBUG

/* 
 * only usable from one interpreter at a time -- copies interpreter because
 * malloc is not normally called with an interpreter pointer
 */

static Tcl_Interp *myinterp;

#define GUARD_SIZE 8

struct mem_header {
        long               length;
        struct mem_header *flink;
        struct mem_header *blink;
        char               low_guard[GUARD_SIZE];
        char               body[1];
};

static struct mem_header *allocHead = NULL;  /* List of allocated structures */

#define GUARD_VALUE  '\341'

static char high_guard[] = {0x89, 0xab, 0xcd, 0xef};

static long total_mallocs = 0;
static long total_frees = 0;
static long current_bytes_malloced = 0;
static long maximum_bytes_malloced = 0;
static long current_malloc_packets = 0;
static long maximum_malloc_packets = 0;
static long break_on_malloc = 0;
static long trace_on_at_malloc = 0;
static int  alloc_tracing = FALSE;
#ifdef MEM_VALIDATE
    static int  validate_memory = TRUE;
#else
    static int  validate_memory = FALSE;
#endif


/*
 *----------------------------------------------------------------------
 *
 * dump_memory_info --
 *     Display the global memory management statistics.
 *
 *----------------------------------------------------------------------
 */
static void
dump_memory_info(outFile) 
    FILE *outFile;
{
        fprintf(outFile,"total mallocs             %10d\n", 
                total_mallocs);
        fprintf(outFile,"total frees               %10d\n", 
                total_frees);
        fprintf(outFile,"current packets allocated %10d\n", 
                current_malloc_packets);
        fprintf(outFile,"current bytes allocated   %10d\n", 
                current_bytes_malloced);
        fprintf(outFile,"maximum packets allocated %10d\n", 
                maximum_malloc_packets);
        fprintf(outFile,"maximum bytes allocated   %10d\n", 
                maximum_bytes_malloced);
}

/*
 *----------------------------------------------------------------------
 *
 * ValidateMemory --
 *     Procedure to validate allocted memory guard zones.
 *
 *----------------------------------------------------------------------
 */
static void
ValidateMemory (memHeaderP, file, line, nukeGuards)
    struct mem_header *memHeaderP;
    char              *file;
    char              *line;
    int                nukeGuards;
{
char *hiPtr;
int i;
int guard_failed = FALSE;

for (i = 0; i < GUARD_SIZE; i++)
  if (*(memHeaderP->low_guard+i) != GUARD_VALUE)
    {
    guard_failed = TRUE;
    fprintf(stderr, "low guard byte %d is 0x%x\n", i, 
            *(memHeaderP->low_guard+i) & 0xff);
    }

  if (guard_failed)
    {
    fflush (stdout);
    dump_memory_info (stderr);
    panic ("ckfree: low guard failed at %lx, %s %d", memHeaderP->body,
           file, line);
    }

hiPtr = (char *)memHeaderP->body + memHeaderP->length;
for (i = 0; i < GUARD_SIZE; i++)
  if (*(hiPtr+i) != GUARD_VALUE)
    {
    fprintf(stderr, "hi guard byte %d is 0x%x\n", i, *(hiPtr+i) & 0xff);
    guard_failed = TRUE;
    }

  if (guard_failed)
    {
    fflush (stdout);
    dump_memory_info (stderr);
    panic ("ckfree: high guard failed at %lx, %s %d", memHeaderP->body, 
           file, line);
    }

if (nukeGuards)
    {
    memset (memHeaderP->low_guard, 0, GUARD_SIZE); 
    memset (hiPtr, 0, GUARD_SIZE); 
    }

}

/*
 *----------------------------------------------------------------------
 *
 * ValidateAllMemory --
 *     Validates guard regions for all allocated memory.
 *
 *----------------------------------------------------------------------
 */
void
ValidateAllMemory (file, line)
    char  *file;
    char  *line;
{
struct mem_header *memScanP;

for (memScanP = allocHead; memScanP != NULL; memScanP = memScanP->flink) {
    ValidateMemory (memScanP, file, line, FALSE);
    }
}

/*
 *----------------------------------------------------------------------
 *
 * db_ckalloc - debugging ckalloc
 *
 *        Allocate the requested amount of space plus some extra for
 *        guard bands at both ends of the request, plus a size, panicing 
 *        if there isn't enough space, then write in the guard bands
 *        and return the address of the space in the middle that the
 *        user asked for.
 *
 *        The second and third arguments are file and line, these contain
 *        the filename and line number corresponding to the caller.
 *        These are sent by the ckalloc macro; it uses the preprocessor
 *        autodefines __FILE__ and __LINE__.
 *
 *----------------------------------------------------------------------
 */
char *db_ckalloc(size, file, line)
    unsigned int size;
    char        *file;
    int          line;
{
        struct mem_header *result;
        extern char *malloc();

        if (validate_memory)
                ValidateAllMemory (file, line);

        result = (struct mem_header *)malloc((unsigned)size + 
                                  sizeof(struct mem_header) + GUARD_SIZE);
        if (result == NULL) {
                fflush(stdout);
                dump_memory_info(stderr);
                panic("unable to alloc %d bytes, %s line %d", size, file, 
                      line);
        }

        /*
         * Fill in guard zones and size.  Link into allocated list.
         */
        result->length = size;
	memset (result->low_guard, GUARD_VALUE, GUARD_SIZE);
        memset (result->body + size, GUARD_VALUE, GUARD_SIZE);
        result->flink = allocHead;
        result->blink = NULL;
        if (allocHead != NULL)
                allocHead->blink = result;
        allocHead = result;

        total_mallocs++;
        if (trace_on_at_malloc && (total_mallocs >= trace_on_at_malloc)) {
                (void) fflush(stdout);
                fprintf(stderr, "reached malloc trace enable point (%d)\n",
                        total_mallocs);
                fflush(stderr);
                alloc_tracing = TRUE;
                trace_on_at_malloc = 0;
        }

        if (alloc_tracing)
                fprintf(stderr,"ckalloc %lx %d %s %d\n", result->body, size, 
                        file, line);

        if (break_on_malloc && (total_mallocs >= break_on_malloc)) {
                break_on_malloc = 0;
                (void) fflush(stdout);
                fprintf(stderr,"reached malloc break limit (%d)\n", 
                        total_mallocs);
                fprintf(stderr, "program will now enter C debugger\n");
                (void) fflush(stderr);
                kill (getpid(), SIGINT);
        }

        current_malloc_packets++;
        if (current_malloc_packets > maximum_malloc_packets)
                maximum_malloc_packets = current_malloc_packets;
        current_bytes_malloced += size;
        if (current_bytes_malloced > maximum_bytes_malloced)
                maximum_bytes_malloced = current_bytes_malloced;

        return result->body;
}

/*
 *----------------------------------------------------------------------
 *
 * db_ckfree - debugging ckfree
 *
 *        Verify that the low and high guards are intact, and if so
 *        then free the buffer else panic.
 *
 *        The guards are erased after being checked to catch duplicate
 *        frees.
 *
 *        The second and third arguments are file and line, these contain
 *        the filename and line number corresponding to the caller.
 *        These are sent by the ckfree macro; it uses the preprocessor
 *        autodefines __FILE__ and __LINE__.
 *
 *----------------------------------------------------------------------
 */
void
db_ckfree(where, file, line)
    char *where;
    char *file;
    int   line;
{
        struct mem_header *memp = 0;  /* Must be zero for size calc */

        /*
         * Since header ptr is zero, body offset will be size
         */
        memp = (struct mem_header *)(where - (int)memp->body);

        if (alloc_tracing)
                fprintf(stderr, "ckfree %lx %ld %s %d\n", memp->body, 
                        memp->length, file, line);
        
        if (validate_memory)
                ValidateAllMemory (file, line);

        ValidateMemory (memp, file, line, TRUE);

        total_frees++;
        current_malloc_packets--;
        current_bytes_malloced -= memp->length;

        /*
         * Delink from allocated list
         */
        if (memp->flink != NULL)
                memp->flink->blink = memp->blink;
        if (memp->blink != NULL)
                memp->blink->flink = memp->flink;
        if (allocHead == memp)
                allocHead = memp->flink;
        free(memp);
}

/*
 *----------------------------------------------------------------------
 *
 * MemoryCmd --
 *     Implements the TCL memory command:
 *       memory info
 *       break_on_malloc count
 *       trace_on_at_malloc count
 *       trace on|off
 *       validate on|off
 *
 * Results:
 *     Standard TCL results.
 *
 *----------------------------------------------------------------------
 */
static int
MemoryCmd (clientData, interp, argc, argv)
    char       *clientData;
    Tcl_Interp *interp;
    int         argc;
    char      **argv;
{
        if (argc < 2)
                return Tcl_RetError (interp, "TCL",
                              "bad # arg: %.50s subcommand [args..]", argv[0]);

        if (strcmp(argv[1],"trace") == 0) {
                if (argc != 3)
                        return Tcl_RetError (interp, "TCL", 
                               "bad suboption: %.50s trace [on|off]", argv[0]);

                alloc_tracing = (strcmp(argv[2],"on") == 0);
                return TCL_OK;
        }

        if (strcmp(argv[1],"validate") == 0) {
                if (argc != 3)
                        return Tcl_RetError (interp, "TCL", 
                            "bad suboption: %.50s validate [on|off]", argv[0]);

                validate_memory = (strcmp(argv[2],"on") == 0);
                return TCL_OK;
        }

        if (strcmp(argv[1],"trace_on_at_malloc") == 0) {

                if (argc != 3) {
                        err:
                        return Tcl_RetError (interp, "TCL",
                          "bad # arg or non-numeric value: %.50s %.50s value",
                           argv[0], argv[1]);
                }
                if (!StrToLong (argv[2], 10, &trace_on_at_malloc))
                        goto err;
                return TCL_OK;
        }

        if (strcmp(argv[1],"break_on_malloc") == 0) {

                if (argc != 3) 
                        goto err;
                if (!StrToLong (argv[2], 10, &break_on_malloc))
                        goto err;
                return TCL_OK;
        }

        if (strcmp(argv[1],"info") == 0) {
                dump_memory_info(stdout);
                return TCL_OK;
        }
        return Tcl_RetError (interp, "TCL", 
                     "%.50s suboption is not info, %s%s",
                      argv[0], "break_on_malloc, trace_on_at_malloc, ",
                      "trace, validate");
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_InitMemory --
 *     Initialize the memory command.
 *
 *----------------------------------------------------------------------
 */
Tcl_InitMemory(interp)
    Tcl_Interp *interp;
{
Tcl_CreateCommand (interp, "memory", MemoryCmd, (ClientData)NULL, 
                  (void (*)())NULL);
}

#else


/*
 *----------------------------------------------------------------------
 *
 * ckalloc --
 *     Interface to malloc when MEM_DEBUG is disabled.  It does check
 *     that memory was actually allocated.
 *
 *----------------------------------------------------------------------
 */
char *
ckalloc(size)
unsigned int size;
{
        char *result;
        extern char *malloc();

        result = malloc(size);
        if (result == NULL) 
                panic("unable to alloc %d bytes", size);
        return result;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_InitMemory --
 *     Dummy initialization for memory command, which is only available 
 *     if MEM_DEBUG is on.
 *
 *----------------------------------------------------------------------
 */
Tcl_InitMemory(interp)
Tcl_Interp *interp;
{
}

#endif
