/* 
 * tclArray.c --
 *
 *	Contains the basic facilities for TCL arrays
 *
 */

/* 
 * Copyright 1990 Karl Lehenbauer, Mark Diekhans, 
 *                Peter da Silva and Jordan Henderson.
 *
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that the above copyright
 * notice appear in all copies.  Karl Lehenbauer, Mark Diekhans,
 * Peter da Silva and Jordan Henderson make no representations
 * about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 */

#ifndef NOSCCSID
	static char *sccs_id = "%M% %I% %H%";
#endif

#include <assert.h>

#if defined(XENIX) || defined(BSD)
#include <stdio.h>	/* $%&*#$% Xenix assert.h requires stdio.h */
#endif

#include "tclInt.h"

#define DEFAULT_BUCKETS 32

extern Var *FindVar();
extern Var *NewVar();

int
calc_hash(key,limit)
char *key;
int limit;
{
    unsigned int result = 0;

    while (*key != '\0')
	result += (*key++ * 101) + 191;

    return(result % limit);
}

AssociativeArray *CreateAssociativeArray(nbuckets)
int nbuckets;
{
    AssociativeArray *ap;
    ArrayEntry *entryp;
    int j;

    ap = (AssociativeArray *)ckalloc(sizeof(struct AssociativeArray));
    ap->nElements = nbuckets;
    ap->entryList = (struct ArrayEntry **)ckalloc(nbuckets * sizeof(struct ArrayEntry *));
    for (j = 0; j < nbuckets; j++)
	ap->entryList[j] = (struct ArrayEntry *)NULL;
    
    return(ap);
}

void
DestroyAssociativeArray(ap)
AssociativeArray *ap;
{
    int i;
    ArrayEntry *entryp;

    for (i = 0; i < ap->nElements; i++)
    {
	entryp = ap->entryList[i];
	while (entryp != NULL)
	{
	    struct ArrayEntry *nextp;

	    if (entryp->key != NULL)
		ckfree(entryp->key);

	    if (entryp->value != NULL)
		ckfree(entryp->value);

	    nextp = entryp->nextEntry;
	    ckfree(entryp);
	    entryp = nextp;
	}
    }
    ckfree(ap->entryList);
    ckfree(ap);
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_FindArrayEntry --
 *
 *	Return the value of a Tcl associative array entry
 *
 * Results:
 *	The return value points to the current value of 
 *      varName(arrayEntry).  If the variable is not defined in interp,
 *      either as a local or global variable, then a NULL pointer is
 *      returned.  Note:  the
 *	return value is only valid up until the next call to Tcl_SetVar;
 *	if you depend on the value lasting longer than that, then make
 *	yourself a private copy.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

struct ArrayEntry *
FindArrayEntry(ap, arrayIndex)
AssociativeArray *ap;
char *arrayIndex;
{
    int index;
    ArrayEntry *entryp;

    index = calc_hash(arrayIndex, ap->nElements);
    entryp = ap->entryList[index];
    if (entryp == NULL)
        return NULL;
    for (; entryp != NULL; entryp = entryp->nextEntry)
    {
	if ((*arrayIndex == *entryp->key) && (strcmp(arrayIndex, entryp->key) == 0))
	{
	    return(entryp);
	}
    }
    return NULL;
}


/*
 *----------------------------------------------------------------------
 *
 * Tcl_GetArrayVar --
 *
 *	Return the value of a Tcl variable.
 *
 * Results:
 *	The return value points to the current value of varName.  If
 *	the variable is not defined in interp, either as a local or
 *	global variable, then a NULL pointer is returned.  Note:  the
 *	return value is only valid up until the next call to 
 *	Tcl_SetArrayVar; if you depend on the value lasting longer than 
 *      that, then make yourself a private copy.
 *
 * Side effects:
 *	None.
 *
 * Special Note:
 *	This code is based on Tcl_SetVar and has a lot of similarities
 *	to it.
 *----------------------------------------------------------------------
 */

char *
Tcl_GetArrayVar(interp, varName, arrayIndex, global)
    Tcl_Interp *interp;		/* Command interpreter in which varName is
				 * to be looked up. */
    char *varName;		/* Name of a variable in interp. */
    char *arrayIndex;		/* name of the index. */
    int global;			/* If non-zero, use only a global variable */
{
    Var *varPtr;
    AssociativeArray *arrayp;
    ArrayEntry *entryp;
    Interp *iPtr = (Interp *) interp;

    if (global || (iPtr->varFramePtr == NULL)) {
	varPtr = FindVar(&iPtr->globalPtr, varName);
    } else {
	varPtr = FindVar(&iPtr->varFramePtr->varPtr, varName);
    }
    if (varPtr == NULL) {
	return NULL;
    }
    if (varPtr->flags & VAR_GLOBAL) {
	varPtr = varPtr->globalPtr;
    }
    if ((arrayp = varPtr->arrayPtr) == NULL) {
	return NULL;
    }
    entryp = FindArrayEntry(arrayp, arrayIndex);
    if (entryp == NULL) {
	return NULL;
    } else {
	return entryp->value;
    }
}


/*
 *----------------------------------------------------------------------
 *
 * Tcl_SetArrayVar --
 *
 *	Change the value of a variable.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	If varName is defined as a local or global variable in interp,
 *	its value is changed to newValue.  If varName isn't currently
 *	defined, then a new global variable by that name is created.
 *
 * Special Note:
 *	This code is based on Tcl_SetVar and has a lot of similarities
 *	to it.
 *
 *----------------------------------------------------------------------
 */

void
Tcl_SetArrayVar(interp, varName, arrayIndex, newValue, global)
    Tcl_Interp *interp;		/* Command interpreter in which varName is
				 * to be looked up. */
    char *varName;		/* Name of a variable in interp. */
    char *arrayIndex;		/* name of the index. */
    char *newValue;		/* New value for varName. */
    int global;			/* If non-zero, use only a global variable. */
{
    register Var *varPtr, **varListPtr;
    register Interp *iPtr = (Interp *) interp;
    int valueLength;
    int index;
    AssociativeArray *arrayp;
    ArrayEntry *entryp, *old_entryp;

    /* locate the variable; create it if not found */
    if (global || (iPtr->varFramePtr == NULL)) {
	varListPtr = &iPtr->globalPtr;
    } else {
	varListPtr = &iPtr->varFramePtr->varPtr;
    }
    varPtr = FindVar(varListPtr, varName);
    if (varPtr == NULL) {
	varPtr = NewVar(varName, "");
        varPtr->flags |= VAR_DOESNT_EXIST;
	varPtr->nextPtr = *varListPtr;
	*varListPtr = varPtr;
    } else {
	if (varPtr->flags & VAR_GLOBAL) {
		varPtr = varPtr->globalPtr;
	}
    }

    /* if the variable doesn't have an array attached, create an array */
    if ((arrayp = varPtr->arrayPtr) == NULL) {
	arrayp = varPtr->arrayPtr = CreateAssociativeArray(DEFAULT_BUCKETS);
    }

    valueLength = strlen(newValue);
    index = calc_hash(arrayIndex, arrayp->nElements);

    /* look up the requested key.  if the entryList (bucket list) entry
     * is NULL, there are no keys in this bucket, so link this new guy to
     * it */
    entryp = arrayp->entryList[index];
    if (entryp == NULL)
    {
	entryp = (struct ArrayEntry *)ckalloc(sizeof(struct ArrayEntry));
	arrayp->entryList[index] = entryp;

    copy_everything:
	entryp->nextEntry = NULL;
	entryp->key = ckalloc(strlen(arrayIndex) + 1);
	strcpy(entryp->key, arrayIndex);

    copy_value:
	entryp->value = ckalloc(valueLength + 1);
	strcpy(entryp->value, newValue);

	entryp->valueLength = valueLength;
	return;
    }

    /* else there is at least one key shooting off from this bucket,
     * follow this buck entry's entry list until matching or coming
     * to the end */
    for (; entryp != NULL; old_entryp = entryp, entryp = entryp->nextEntry)
    {
	if ((*arrayIndex == *entryp->key) && (strcmp(arrayIndex, entryp->key) == 0))
	{
	    if (valueLength > entryp->valueLength)
	    {
		ckfree(entryp->value);
		goto copy_value;
	    } else {
		strcpy(entryp->value, newValue);
		return;
	    }
	}
    }
    old_entryp->nextEntry = entryp = (struct ArrayEntry *)ckalloc(sizeof(struct ArrayEntry));
    goto copy_everything;
}


/*
 *----------------------------------------------------------------------
 *
 * Tcl_NextElementCmd --
 *
 *	This procedure is invoked to process the "next_element" Tcl
 *      command.  
 *
 * Results:
 *	A standard Tcl result is the value of the next element
 *      in the specified associative array.  If the specified variable
 *	is empty, returns the first element.  If the second form
 *	is used, the key returned is assigned to a second variable
 *	and a one is returned if next_element succeeded, else a zero.
 *
 * Side effects:
 *	Sets the specified variable to contain a "magic cookie"
 *	that can be used on a subsequent call to find the next
 *	element.  The variable is set to be empty on the call
 *	subsequent to the one that returned the last element
 *	in the array.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
int
Tcl_NextElementCmd(dummy, interp, argc, argv)
    ClientData dummy;			/* Not used. */
    Interp *interp;			/* Current interpreter. */
    int argc;				/* Number of arguments. */
    char **argv;			/* Argument strings. */
{
    char *array_element_cookie;
    Var *varPtr;
    AssociativeArray *arrayp;
    ArrayEntry *entryp;
    int bucket_index, entry_index;
    Interp *iPtr = (Interp *) interp;

    if (argc != 4) 
        return Tcl_RetError (iPtr, "TCL", 
           "bad # arg: %.50s array_name var_name result_key_varname", argv[0]);

    array_element_cookie = Tcl_GetVar(iPtr, argv[2], 0);
    if (array_element_cookie == NULL) 
	return Tcl_RetError (iPtr, "TCL", "variable or array element \"%s\" doesn't exist", 
                             argv[2]);

    if (*array_element_cookie == '\0') {
	bucket_index = -1;
	entry_index = 0;
    } else {
	sscanf(array_element_cookie, "%d:%d", &bucket_index, &entry_index);
    }
    if (iPtr->varFramePtr == NULL) {
	varPtr = FindVar(&iPtr->globalPtr, argv[1]);
    } else {
	varPtr = FindVar(&iPtr->varFramePtr->varPtr, argv[1]);
    }
    if (varPtr == NULL) {
	no_exist:
	return Tcl_RetError (iPtr, "TCL", "%.50s: array \"%s\" doesn't exist", 
                             argv[0], argv[1]);
    }
    if (varPtr->flags & VAR_GLOBAL) {
	varPtr = varPtr->globalPtr;
    }
    if ((arrayp = varPtr->arrayPtr) == NULL) goto no_exist;

    if (bucket_index == -1) {
	entryp = NULL;
    } else {
	int j;

	entryp = arrayp->entryList[bucket_index];
	assert(entryp != NULL);
	for (j = 0; j < entry_index; j++) {
	    assert(entryp->nextEntry);
	    entryp = entryp->nextEntry;
	}
    }

    if (entryp != NULL) {
	entryp = entryp->nextEntry;
	if (entryp) {
	    entry_index++;
	    goto done;
	}
    }
    entry_index = 0;

    for (bucket_index = bucket_index + 1; bucket_index < arrayp->nElements; bucket_index++) {
	if (entryp = arrayp->entryList[bucket_index]) {
	    entry_index = 0;
	    goto done;
	}
    }
    /* no more entries, clear cookie var; clear key var & return 0 if
     * second var was specified */
    Tcl_SetVar(iPtr, argv[2], "", 0);
    Tcl_SetVar(iPtr, argv[3], "", 0);
    Tcl_Return(iPtr, "0", TCL_STATIC);
    return TCL_OK;

    /* found an entry, store the new cookie. */
    done:
    sprintf(iPtr->result, "%d:%d", bucket_index, entry_index);
    Tcl_SetVar(iPtr, argv[2], iPtr->result, 0);
    /* return the key or, if it's the two-variable form, set the
     * second specified variable to be equal to the key and return
     * 1 for success */
    Tcl_SetVar(iPtr, argv[3], entryp->key, 0);
    Tcl_Return(iPtr, "1", TCL_STATIC);
    return TCL_OK;
}


/*
 *----------------------------------------------------------------------
 *
 * Tcl_UnsetArrayVar --
 *
 *	Delete an associative array entry
 *
 * Results:
 *	None.
 *
 *----------------------------------------------------------------------
 */

int
Tcl_UnsetArrayVar(interp, varName, arrayIndex, global)
    Tcl_Interp *interp;		/* Command interpreter in which varName is
				 * to be looked up. */
    char *varName;		/* Name of a variable in interp. */
    char *arrayIndex;		/* name of the index. */
    int global;			/* If non-zero, use only a global variable. */
{
    register Var *varPtr, **varListPtr;
    register Interp *iPtr = (Interp *) interp;
    int index;
    AssociativeArray *arrayp;
    ArrayEntry *entryp, *old_entryp = NULL;

    /* locate the variable; error it if not found */
    if (global || (iPtr->varFramePtr == NULL)) {
	varListPtr = &iPtr->globalPtr;
    } else {
	varListPtr = &iPtr->varFramePtr->varPtr;
    }
    varPtr = FindVar(varListPtr, varName);
    if (varPtr == NULL) {
	return TCL_ERROR;
    } else {
	if (varPtr->flags & VAR_GLOBAL) {
		varPtr = varPtr->globalPtr;
	}
    }

    /* if the variable doesn't have an array attached, return */
    if ((arrayp = varPtr->arrayPtr) == NULL) {
	return TCL_ERROR;
    }

    index = calc_hash(arrayIndex, arrayp->nElements);

    /* look up the requested key.  if the entryList (bucket list) entry
     * is NULL, there are no keys in this bucket */
    entryp = arrayp->entryList[index];
    if (entryp == NULL)
	goto notFound;
    
    /* else there is at least one key shooting off from this bucket,
     * follow this buck entry's entry list until matching or coming
     * to the end */
    for (; entryp != NULL; old_entryp = entryp, entryp = entryp->nextEntry)
    {
	if ((*arrayIndex == *entryp->key) && (strcmp(arrayIndex, entryp->key) == 0))
	{
	    if (old_entryp) {
		old_entryp->nextEntry = entryp->nextEntry;
	    } else {
		arrayp->entryList[index] = entryp->nextEntry;
	    }
	    ckfree(entryp->key);
	    ckfree(entryp->value);
	    ckfree(entryp);
	    return TCL_OK;
	}
    }
    notFound:
    return Tcl_RetError (interp, "TCL", "couldn't find array element");
}

