/* 
 * Copyright 1990 Karl Lehenbauer, Mark Diekhans, 
 *                Peter da Silva and Jordan Henderson.
 *
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that the above copyright
 * notice appear in all copies.  Karl Lehenbauer, Mark Diekhans,
 * Peter da Silva and Jordan Henderson make no representations
 * about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 */

#ifndef NOSCCSID
        static char *sccs_id = "%M% %I% %H%";
#endif

#include <stdio.h>
#include <sys/types.h>
#include <sys/times.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "tcl.h"

#include "iocmds.h"
extern FILE *popen();

struct handle handles[MAXFILES];


/*
 *----------------------------------------------------------------------
 *
 * convfile --
 *
 *      Convert a textual filehandle (passed as an argument to a
 *      Unix I/O function) of the form "filen" where n is an
 *      integer, or "stdin", "stdout" or "stderr" to a file 
 *      descriptor.
 *
 * Results:
 *
 *   Returns TCL_OK is conversion succeeded, returns TCL_ERROR and
 *   error text in interp->result if conversion did not succeed.
 *   The "i" argument is a pointer to int that is set to the file
 *   descriptor specified in the handle if conversion succeeded.
 *
 *----------------------------------------------------------------------
 */
convfile (interp, modname, s, i, opened)
    Tcl_Interp *interp;
    char       *modname;
    char       *s;
    int        *i;
    int         opened;
{
    int   result = -1;

    /* if the first character of the specified filehandle is 's', check
     * it for the aliases stdin, stdout and stderr.  Otherwise it must
     * be of the form "filen" where n is an integer */
    if (*s == 's')
        {
        if (strncmp(s, "stdout", 6) == 0)
                result = 1;
        else if (strncmp(s, "stdin", 5) == 0)
                result = 0;
        else if (strncmp(s, "stderr", 6) == 0)
                result = 2;
        } 
    else
        if (strncmp(s, "file", 4) == 0)
            if (!StrToInt (&s[4], 10, &result))
                result = -1;  /* Remember its bad */
    if (result < 0)
        return Tcl_RetError (interp, "TCL", "%s: badly formed file handle", 
                             modname);

    if (result >= MAXFILES)
        return Tcl_RetError (interp, "TCL", "%s: file ID out of range", 
                             modname);

    *i = (int)result;
    if (opened & (handles[result].fp == NULL))
            return Tcl_RetError (interp, "TCL", "%.50s: file not open", 
                                 modname);

    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * DoClose --
 *     Close the specified file or pipe.
 *
 * Results:
 *     TCL_OK if it was closed OK, TCL_ERROR and a error message in the
 *     interpreter result if close failed.
 *
 *----------------------------------------------------------------------
 */
int
DoClose (interp, fileIdx)
    Tcl_Interp *interp;
    int         fileIdx;
{
	int result;

	if (handles[fileIdx].flags & PIPE_HANDLE)
            result = pclose(handles[fileIdx].fp);
        else
            result = fclose(handles[fileIdx].fp);
        handles[fileIdx].fp = NULL;
	if (result == EOF)
	    return Tcl_RetSysError (interp, (char *)NULL);
        return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * CmdOpen --
 *     Implements the open TCL command:
 *         open "filename access"
 *     See the iocmds(TCL) manual page.
 *
 * Results:
 *      Returns TCL_OK and interp->result containing a filehandle
 *      if the requested file or pipe was successfully opened.
 *
 *      Return TCL_ERROR and interp->result containing an
 *      explanation of what went wrong if an error occured.
 *
 * Side effects:
 *      Locates and creates an entry in the handles table
 *
 *----------------------------------------------------------------------
 */
int
CmdOpen (clientData, interp, argc, argv)
    char       *clientData;
    Tcl_Interp *interp;
    int         argc;
    char      **argv;
{
    int   idx;
    char *fileName;

    if (argc != 3)
        return Tcl_RetError (interp, "TCL", "bad # arg: %.50s filename rwa+", 
                             argv[0]);

    for (idx = 0; idx < MAXFILES; idx++)
        if (handles[idx].fp == NULL)
            break;

    if (idx == MAXFILES)
        return Tcl_RetError (interp, "TCL", "%.50s: too many open files", 
                             argv[0]);

    if (argv[1][0] != '|')
    {
        fileName = argv[1];
        if (*fileName == '~')
            if ((fileName = Tcl_TildeSubst (interp, fileName)) == NULL)
                return TCL_ERROR;
              
        handles[idx].fp = fopen(fileName,argv[2]);
        handles[idx].flags = 0;
    }
    else
    {
        handles[idx].fp = popen(&argv[1][1],argv[2]);
        handles[idx].flags = PIPE_HANDLE;
    }
    if (handles[idx].fp == NULL)
        return Tcl_RetSysError (interp, "%s", argv[1]);

    strncpy (handles[idx].mode, argv[2], 2);
    handles[idx].mode[2] = '\0';
    sprintf(interp->result,"file%d",idx);
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * CmdClose --
 *     Implements the close TCL command:
 *        close [filehandle1...filehandleN]
 *     See the iocmds(TCL) manual page.
 *
 * Results:
 *      Closes specified files or all if "all" keyword is specified.
 *      Returns TCL_ERROR with interp->result containing an explanation
 *      of the error if something went wrong, else returns TCL_OK
 *      with interp->result empty.
 *
 * Side effects:
 *      clears the file's entry in the handles table
 *
 *----------------------------------------------------------------------
 */
int
CmdClose (clientData, interp, argc, argv)
    char       *clientData;
    Tcl_Interp *interp;
    int         argc;
    char      **argv;
{
    int i;
    int filen;

    if (argc < 2)
        return Tcl_RetError (interp, "TCL", 
                          "bad # arg: %.50s filen [filen..] | [all]", argv[0]);

    if ((argc == 2) && (strcmp(argv[1],"all") == 0))
        {
        for (i = 3; i < MAXFILES; i++)
            {
            if (handles[i].fp != NULL)
                if (DoClose (interp, i) != TCL_OK)
                    return TCL_ERROR;
            }
        return TCL_OK;
        }

    for (i = 1; i < argc; i++)
        {
        if (convfile(interp, argv[0], argv[i], &filen, 1) == TCL_ERROR)
            return TCL_ERROR;

        if (handles[filen].fp != NULL)
            if (DoClose (interp, filen) != TCL_OK)
                 return TCL_ERROR;
        }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * CmdDup --
 *     Implements the dup TCL command:
 *         dup filehandle
 *     See the iocmds(TCL) manual page.
 *
 * Results:
 *      Returns TCL_OK and interp->result containing a filehandle
 *      if the requested file or pipe was successfully duplicated.
 *
 *      Return TCL_ERROR and interp->result containing an
 *      explanation of what went wrong if an error occured.
 *
 * Side effects:
 *      Locates and creates an entry in the handles table
 *
 *----------------------------------------------------------------------
 */
int
CmdDup (clientData, interp, argc, argv)
    char       *clientData;
    Tcl_Interp *interp;
    int         argc;
    char      **argv;
{
        int newIdx, oldIdx, fileId;

        if (argc != 2)
                return Tcl_RetError (interp, "TCL", 
                                    "bad # arg: %.50s filehandle", argv[0]);

        for (newIdx = 0; newIdx < MAXFILES; newIdx++)
                if (handles[newIdx].fp == NULL) break;

        if (newIdx == MAXFILES)
                return Tcl_RetError (interp, "TCL", 
                                     "%.50s: too many open files", argv[0]);

        if (convfile(interp, argv[0], argv[1], &oldIdx, 1) == TCL_ERROR)
            return TCL_ERROR;

	if ((fileId = dup (fileno (handles[oldIdx].fp))) < 0 )
                return Tcl_RetSysError (interp, "%s", argv[0]);

        if ((handles[newIdx].fp = fdopen (fileId, handles[oldIdx].mode)) == 
             NULL)
                return Tcl_RetSysError (interp, "%s", argv[0]);

        handles[newIdx].flags = handles[oldIdx].flags;
        strcpy (handles[newIdx].mode, handles[oldIdx].mode);

        sprintf(interp->result,"file%d",newIdx);
        return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * CmdDup2 --
 *     Implements the dup2 TCL command:
 *         dup2 oldhandle newhandle
 *     See the iocmds(TCL) manual page.
 *
 * Results:
 *      Returns TCL_OK and interp->result containing a filehandle
 *      if the requested file or pipe was successfully duplicated.
 *
 *      Return TCL_ERROR and interp->result containing an
 *      explanation of what went wrong if an error occured.
 *
 * Side effects:
 *      Locates and creates an entry in the handles table
 *
 *----------------------------------------------------------------------
 */
int
CmdDup2 (clientData, interp, argc, argv)
    char       *clientData;
    Tcl_Interp *interp;
    int         argc;
    char      **argv;
{
        int newIdx, oldIdx, fileId;

        if (argc != 3)
                return Tcl_RetError (interp, "TCL", 
                             "bad # arg: %.50s oldhandle newhandle", argv[0]);

        if (convfile(interp, argv[0], argv[1], &oldIdx, 1) == TCL_ERROR)
            return TCL_ERROR;

        if (convfile(interp, argv[0], argv[2], &newIdx, 0) == TCL_ERROR)
            return TCL_ERROR;
        
        if (handles[newIdx].fp != NULL)
	    if (DoClose (interp, newIdx) != TCL_OK)
                return TCL_ERROR;

	if ((fileId = dup (fileno (handles[oldIdx].fp))) < 0 )
                return Tcl_RetSysError (interp, "%s", argv[0]);

        if ((handles[newIdx].fp = fdopen (fileId, handles[oldIdx].mode)) ==
             NULL)
                return Tcl_RetSysError (interp, "%s", argv[0]);

        handles[newIdx].flags = handles[oldIdx].flags;
        strcpy (handles[newIdx].mode, handles[oldIdx].mode);

        return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * CmdPipe --
 *     Implements the pipe TCL command:
 *         pipe [handle_var_r handle_var_w]
 *     See the iocmds(TCL) manual page.
 *
 * Results:
 *      Standard TCL result.
 *
 * Side effects:
 *      Locates and creates entries in the handles table
 *
 *----------------------------------------------------------------------
 */
int
CmdPipe (clientData, interp, argc, argv)
    char       *clientData;
    Tcl_Interp *interp;
    int         argc;
    char      **argv;
{
        int  fhIdx1, fhIdx2;
        int  fileIds [2];
        char fHandle1 [10];
        char fHandle2 [10];

        if (!((argc == 1) || (argc == 3)))
                return Tcl_RetError (interp, "TCL", 
                      "bad # arg: %.50s [handle_var_r handle_var_w", argv[0]);

        for (fhIdx1 = 0; fhIdx1 < MAXFILES; fhIdx1++)
                if (handles[fhIdx1].fp == NULL) break;
        for (fhIdx2 = fhIdx1+1; fhIdx2 < MAXFILES; fhIdx2++)
                if (handles[fhIdx2].fp == NULL) break;

        if (fhIdx2 >= MAXFILES)
                return Tcl_RetError (interp, "TCL", 
                                     "%.50s: too many open files", argv[0]);

        if (pipe (fileIds) < 0)
                return Tcl_RetSysError (interp, "%s", argv[0]);

        if ((handles[fhIdx1].fp = fdopen (fileIds[0], "r")) == NULL)
                return Tcl_RetSysError (interp, "%s", argv[0]);
        strcpy (handles[fhIdx1].mode, "r");
        sprintf (fHandle1, "file%d", fhIdx1);

        if ((handles[fhIdx2].fp = fdopen (fileIds[1], "w")) == NULL)
                return Tcl_RetSysError (interp, "%s", argv[0]);
        strcpy (handles[fhIdx2].mode, "w");
        sprintf (fHandle2, "file%d", fhIdx2);
        
        if (argc == 1)      
                sprintf (interp->result,"%s %s", fHandle1, fHandle2);
        else {
                Tcl_SetVar (interp, argv[1], fHandle1, 0);
                Tcl_SetVar (interp, argv[2], fHandle2, 0);
        }
                
        return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * CmdGets --
 *     Implements the gets TCL command:
 *        gets filehandle [var]
 *     See the iocmds(TCL) manual page.
 *
 * Results:
 *      If optional variable is specified, a line is read from filehandle 
 *      into var and the the number of characters read are returned, or 
 *      -1 if EOF
 *
 *      If the optional variable is not specified, return the line read
 *      and the caller is expected to use the eof command to determine
 *      the file status.
 *
 *      An error is returned through TCL_ERROR if an error occurs
 *      while performing the fgets.
 *
 *----------------------------------------------------------------------
 */
int
CmdGets (clientData, interp, argc, argv)
    char       *clientData;
    Tcl_Interp *interp;
    int         argc;
    char      **argv;
{
        int filen;
        FILE *stream;
        char *fail;
        char buf[1024];
        int buflen;

        if ((argc != 2) && (argc != 3))
                return Tcl_RetError (interp, "TCL", 
                         "bad # arg: %.50s filehandle [resultvar]", argv[0]);

        if (convfile(interp, argv[0], argv[1], &filen, 1) == TCL_ERROR)
                return TCL_ERROR;
        stream = handles[filen].fp;
        
        /* Must check if we have read access, stdio does not! */

        if (!((stream->_flag & _IOREAD) || (stream->_flag & _IORW)))
                return Tcl_RetError (interp, "TCL", 
                                     "Not open for read access");

        buf [0] = 0;  /* In case nothing is returned, i.e. EOF */
        fgets(buf, sizeof(buf), stream);
        if (ferror(stream))
        {
                Tcl_RetSysError (interp, (char *)NULL);
                clearerr(stream);
                return TCL_ERROR;
        }
        buf[(buflen = strlen(buf) - 1)] = '\0';
        if (argc == 3)
        {
                int result;

                result = feof(stream) ? 0 : 1;
                Tcl_SetVar(interp, argv[2], buf, 0);
                sprintf(interp->result,"%d", result);
        }
        else
        {
                if (!feof(stream))
                        Tcl_Return(interp, buf, TCL_VOLATILE);
        }

        return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * CmdPuts --
 *     Implements the puts TCL command:
 *         puts filehandle text [no_nl]
 *     See the iocmds(TCL) manual page.
 *
 * Results:
 *      Nothing if OK.
 *
 *      An error is returned through TCL_ERROR if an error occurs
 *      while performing the fgets.
 *
 *----------------------------------------------------------------------
 */
int
CmdPuts (clientData, interp, argc, argv)
    char       *clientData;
    Tcl_Interp *interp;
    int         argc;
    char      **argv;
{
        int filen;
        FILE *stream;

        if ((argc < 3) || (argc > 4))
                return Tcl_RetError (interp, "TCL", 
                       "bad # arg: %.50s filehandle string [no_nl]", argv[0]);
        if ((argc == 4) && (strcmp (argv[3], "no_nl") != 0))
                return Tcl_RetError (interp, "TCL", 
                       "%.50s: invalid option, \"no_nl\" expected", argv[0]);

        if (convfile(interp, argv[0], argv[1], &filen, 1) == TCL_ERROR)
                return TCL_ERROR;
        stream = handles[filen].fp;

        /* Must check if we have write access, stdio does not! */

        if (!((stream->_flag & _IOWRT) || (stream->_flag & _IORW)))
                return Tcl_RetError (interp, "TCL", 
                                     "Not open for write access");

        fputs(argv[2], stream);
        if (argc == 3)
                fputc('\n', stream);
        if (ferror(stream))
        {
                Tcl_RetSysError (interp, (char *)NULL);
                clearerr(stream);
                return TCL_ERROR;
        }
        return TCL_OK;
}


/*
 *----------------------------------------------------------------------
 *
 * CmdSeek --
 *     Implements the seek TCL command:
 *         seek filehandle location [whence]
 *     See the iocmds(TCL) manual page.
 *
 * Results:
 *      If all is well, fseeks the specified file to a given position
 *      and returns that position as an integer file pointer location.
 *
 *      Various errors conditions are detected and returned via
 *      TCL_ERROR.
 *
 *----------------------------------------------------------------------
 */
int
CmdSeek (clientData, interp, argc, argv)
    char       *clientData;
    Tcl_Interp *interp;
    int         argc;
    char      **argv;
{
    int filen;
    FILE *stream;
    long offset;
    long origin = 0;

    if ((argc != 3) && (argc != 4))
        return Tcl_RetError (interp, "TCL", 
             "bad # arg: %.50s filehandle location [origin]", argv[0]);


    if (convfile(interp, argv[0], argv[1], &filen, 1) == TCL_ERROR)
        return TCL_ERROR;
    stream = handles[filen].fp;

    if (!StrToLong (argv[2], 10, &offset))
        return Tcl_RetError (interp, "TCL", "%.50s: non-numeric location",
                     argv[0]);

    /*
     * If no origin is provided, it is implicitly 0 (seek from beginning
     * of file)
     */
    if (argc == 4)
        {
        if (strcmp (argv[3], "start") == 0)
            origin = 0;
        else if (strcmp (argv[3], "cur") == 0)
            origin = 1;
        else if (strcmp (argv[3], "end") == 0)
            origin = 2;
        else
            return Tcl_RetError (interp, "TCL", 
                     "%s: origin must be \"start\", \"cur\" or \"end\"", 
                     argv[0]);
        }
    if (fseek(stream, offset, (int)origin) != 0)
        return Tcl_RetSysError (interp, (char *)NULL);
    *interp->result = 0;
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * CmdTell --
 *     Implements the tell TCL command:
 *         tell filehandle
 *     See the iocmds(TCL) manual page.
 *
 * Results:
 *      If all is well, returns the file position of the filehandle
 *      passed as an argument.
 *
 *----------------------------------------------------------------------
 */
int
CmdTell (clientData, interp, argc, argv)
    char       *clientData;
    Tcl_Interp *interp;
    int         argc;
    char      **argv;
{
        int   filen;
        long  FileLoc;
        FILE *stream;

        if (argc != 2)
                return Tcl_RetError (interp, "TCL", 
                                     "bad # arg: %.50s filehandle", argv[0]);

        if (convfile(interp, argv[0], argv[1], &filen, 1) == TCL_ERROR)
                return TCL_ERROR;
        stream = handles[filen].fp;

        if ((FileLoc = ftell(stream)) < 0)
                return Tcl_RetSysError (interp, (char *)NULL);
        sprintf (interp->result, "%ld", FileLoc);
        return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * CmdEof --
 *      report whether or not the specified filehandle is at end of file
 *        eof filehandle
 *        Returns a 1 if filehandle is at EOF, else it returns a zero.
 *
 * Results:
 *      Returns "1" if the filehandle is at end of file, "0" if it
 *      is not.  
 *
 *----------------------------------------------------------------------
 */
int
CmdEof (clientData, interp, argc, argv)
    char       *clientData;
    Tcl_Interp *interp;
    int         argc;
    char      **argv;
{
        int filen;
        FILE *stream;

        if (argc != 2)
                return Tcl_RetError (interp, "TCL", 
                                     "bad # arg: %.50s filehandle", argv[0]);

        if (convfile(interp, argv[0], argv[1], &filen, 1) == TCL_ERROR)
                return TCL_ERROR;
        stream = handles[filen].fp;

        sprintf (interp->result,"%d", !!feof(stream));
        return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * CmdCopy_open_files --
 *     Implements the copy_open_files TCL command:
 *         copy_open_files handle1 handle2
 *     See the iocmds(TCL) manual page.
 *
 * Results:
 *      Nothing if it worked, else an error.
 *
 *      Both files are left open and located at end-of-file if
 *      the routine completed successfully
 *
 *----------------------------------------------------------------------
 */
int
CmdCopy_open_files(clientData, interp, argc, argv)
    char       *clientData;
    Tcl_Interp *interp;
    int         argc;
    char      **argv;
{
        int fromfilen, tofilen;
        FILE *fromstream, *tostream;
        char transfer_buffer[512];

        if (argc != 3)
                return Tcl_RetError (interp, "TCL", 
                      "bad # arg: %.50s fromfilehandle tofilehandle", argv[0]);

        if (convfile(interp, argv[0], argv[1], &fromfilen, 1) == TCL_ERROR)
                return TCL_ERROR;
        fromstream = handles[fromfilen].fp;

        if (convfile(interp, argv[0], argv[2], &tofilen, 1) == TCL_ERROR)
                return TCL_ERROR;
        tostream = handles[tofilen].fp;

        /* Must check if we have read access, stdio does not! */

        if (!((fromstream->_flag & _IOREAD) || (fromstream->_flag & _IORW)))
                return Tcl_RetError (interp, "TCL", 
                                     "Not open for read access");

        /* Must check if we have write access, stdio does not! */

        if (!((tostream->_flag & _IOWRT) || (tostream->_flag & _IORW)))
                return Tcl_RetError (interp, "TCL", 
                                     "Not open for write access");

        while (1)
                {
                if (fgets (transfer_buffer, sizeof(transfer_buffer),
                           fromstream) == NULL)
                        {
                        if (feof (fromstream))
                                break;
                        else
                                return Tcl_RetSysError (interp, "%s", argv[1]);
                        }
                if (fputs(transfer_buffer, tostream) == EOF)
                        return Tcl_RetSysError (interp, "%s", argv[2]);
                }

        return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * CmdRead --
 *     Implements the read TCL command:
 *         read handle [length]
 *     See the iocmds(TCL) manual page.
 *
 * Results:
 *      The string read if it worked, else an error.
 *
 *----------------------------------------------------------------------
 */
int
CmdRead (clientData, interp, argc, argv)
    char       *clientData;
    Tcl_Interp *interp;
    int         argc;
    char      **argv;
{
int          filen;
FILE        *stream;
struct stat  statbuf;
int          size;

if ((argc != 2) && (argc != 3))
    return Tcl_RetError (interp, "TCL", 
                         "bad # arg: %.50s filehandle [size]",argv[0]);

if (convfile(interp, argv[0], argv[1], &filen, 1) == TCL_ERROR)
    return TCL_ERROR;
stream = handles[filen].fp;

if (argc == 3)
    {
    if (!StrToInt (argv[2], 10, &size))
        return Tcl_RetError (interp, "TCL", "%.50s: non-numeric size", 
                             argv[0]);
    }
else
    {
    if (fstat(fileno (stream), &statbuf) < 0)
        return Tcl_RetSysError (interp, "%s", argv[1]);
    size = statbuf.st_size - ftell(stream);
    }

if (size >= TCL_RESULT_SIZE)
    {
    interp->result = ckalloc(size + 1);
    interp->dynamic = TCL_DYNAMIC;
    }

/* Must check if we have read access, stdio does not! */

if (!((stream->_flag & _IOREAD) || (stream->_flag & _IORW)))
        return Tcl_RetError (interp, "TCL", "Not open for read access");

size = fread(interp->result, 1, size, stream);
if (ferror (stream))
    return Tcl_RetSysError (interp, (char *)NULL);

interp->result[size] = '\0';

return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * CmdFlush --
 *     Implements the flush TCL command:
 *         flush filehandle
 *     See the iocmds(TCL) manual page.
 *
 *----------------------------------------------------------------------
 */
int
CmdFlush (clientData, interp, argc, argv)
    char       *clientData;
    Tcl_Interp *interp;
    int         argc;
    char      **argv;
{
        int   idx;
        int   filen;

        if (argc < 2)
            return Tcl_RetError (interp, "TCL", 
                                 "bad # arg: %.50s filen [filen..]", argv[0]);

        for (idx = 1; idx < argc; idx++)
            {
            if (convfile(interp, argv[0], argv[idx], &filen, 1) == TCL_ERROR)
                return TCL_ERROR;

            if (fflush(handles[filen].fp) == EOF)
                return Tcl_RetSysError (interp, "%s", argv[0]);
            }
        return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_InitIoCmds --
 *      Initialize file I/O commands and install them into Tcl
 *
 * Side effects:
 *      The Unix I/O commands are installed into the specified
 *      interpreter.
 *
 *----------------------------------------------------------------------
 */
Tcl_InitIoCmds(interp)
Tcl_Interp *interp;
{
int i;

/*
 * Initialize the file handle table. The mode is set to append for stdout and
 * stderr in case they are duplicated.
 */
handles[0].fp = stdin;
strcpy (handles[0].mode, "r");
handles[0].flags = 0;

handles[1].fp = stdout;
strcpy (handles[1].mode, "a");
handles[1].flags = 0;

handles[2].fp = stderr;
strcpy (handles[2].mode, "a");
handles[2].flags = 0;

for (i = 3; i < MAXFILES; i++)
    handles[i].fp = NULL;


Tcl_CreateCommand(interp, "open", CmdOpen, (ClientData)NULL, (void (*)())NULL);
Tcl_CreateCommand(interp, "dup",  CmdDup, (ClientData)NULL, (void (*)())NULL);
Tcl_CreateCommand(interp, "dup2", CmdDup2, (ClientData)NULL, (void (*)())NULL);
Tcl_CreateCommand(interp, "pipe", CmdPipe, (ClientData)NULL, (void (*)())NULL);
Tcl_CreateCommand(interp, "close", CmdClose, (ClientData)NULL,
                 (void (*)())NULL);
Tcl_CreateCommand(interp, "gets", CmdGets, (ClientData)NULL, (void (*)())NULL);
Tcl_CreateCommand(interp, "puts", CmdPuts, (ClientData)NULL, (void (*)())NULL);
Tcl_CreateCommand(interp, "seek", CmdSeek, (ClientData)NULL, (void (*)())NULL);
Tcl_CreateCommand(interp, "tell", CmdTell, (ClientData)NULL, (void (*)())NULL);
Tcl_CreateCommand(interp, "eof", CmdEof, (ClientData)NULL, (void (*)())NULL);

Tcl_CreateCommand(interp, "copy_open_files", CmdCopy_open_files,
                 (ClientData)NULL, (void (*)())NULL);
Tcl_CreateCommand(interp, "read", CmdRead, (ClientData)NULL, (void (*)())NULL);
Tcl_CreateCommand(interp, "flush", CmdFlush, (ClientData)NULL,
                 (void (*)())NULL);
}

