/* 
 * Copyright 1990 Karl Lehenbauer, Mark Diekhans, 
 *                Peter da Silva and Jordan Henderson.
 *
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that the above copyright
 * notice appear in all copies.  Karl Lehenbauer, Mark Diekhans,
 * Peter da Silva and Jordan Henderson make no representations
 * about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 */

#include <stdio.h>
#include <errno.h>
#ifdef BSD
#include <sys/time.h>
#endif
#include "tcl.h"

#include "config.h"
#include "patchlevel.h"

char *custom_name = CUSTOM_NAME;
char *custom_longname = CUSTOM_LONGNAME;
int   patchlevel = PATCHLEVEL;
char *tcl_defaultfile = TCL_DEFAULTFILE;
char *tcl_version = TCL_VERSION;


/* initial_commands, this array of pointers to character strings is
 * read through and passed to Tcl_AssembleCmd on startup -- it gets
 * Tcl rolling, the defaults read and the init file pulled in.
 */

static char *initial_commands[] = 
{
"proc set_default {var value} {\n",
"	global $var\n",
"	if ![getenv $var $var] {set $var $value}\n",
"}\n",

"if {[getenv TCLDEFAULT[info tclversion] TCLDEFAULT] || [getenv TCLDEFAULT TCLDEFAULT]} {\n",
"  if {[strequ [cindex $TCLDEFAULT 0] /] || [strequ [cindex $TCLDEFAULT 0] ~] || [strequ [cindex $TCLDEFAULT 0] .]} {\n",
"    if ![file $TCLDEFAULT readable] {\n",
"      puts stderr \"can't find $TCLDEFAULT! (specified by TCLDEFAULT environment var)\"\n",
"      exit 1\n",
"    }\n",
"    source $TCLDEFAULT\n",
"  } else {\n",
"    if {[catch {eval $TCLDEFAULT}] != 0} {\n",
"      puts stderr \"error in Tcl startup code within TCLDEFAULT environment var!\"\n",
"      puts stderr $errorInfo\n",
"      exit 1\n",
"    }\n",
"  }\n",
"} else {\n",
"  set TCLDEFAULT TclDefaults\n",
"  if ![file $TCLDEFAULT readable] {\n",
"    set TCLDEFAULT [custom_info defaultfile][info tclversion]\n",
"    if ![file $TCLDEFAULT readable] {\n",
"      set TCLDEFAULT [custom_info defaultfile]\n",
"      if ![file $TCLDEFAULT readable] {\n",
"        puts stderr \"can't find Tcl default file TclDefaults, $TCLDEFAULT[info tclversion], or\"\n",
"        puts stderr \"$TCLDEFAULT and TCLDEFAULT environment var isn't set!\"\n",
"        exit 1\n",
"      }\n",
"    }\n",
"  }\n",
"  source $TCLDEFAULT\n",
"}\n",
"if ![info exists TCLINIT] {\n",
"  puts stderr \"Tcl default startup file or env-var didn't set TCLINIT Tcl variable!\"\n",
"  exit 1\n",
"}\n",
"source $TCLINIT\n",

 NULL
};


/*
 *----------------------------------------------------------------------
 *
 * Tcl_Startup --
 *
 *      Feed the startup Tcl source code directly to the
 *	  interpreter
 *
 * Results:
 *
 *      Takes an interpreter, and executes everything
 *      it reads from the initial_commands array.
 *
 * Side effects:
 *
 *      Anything that happens as a result of input to eval.
 *      Hopefully, Tcl will get rolling.
 *
 *----------------------------------------------------------------------
 */
int
Eval_Tcl_Startup (interp)
    Tcl_Interp *interp;
{
    Tcl_CmdBuf cmdBuf;
    char       s[256];
    int        result;
    char      *cmd;
    char      *lineptr;
    int        line;

    cmdBuf = Tcl_CreateCmdBuf();

    for (line = 0; ((lineptr = initial_commands[line]) != NULL); line++)
    {
        cmd = Tcl_AssembleCmd(cmdBuf, lineptr);
        if (cmd != NULL)
            {
            result = Tcl_Eval(interp, cmd, 0, (char **)NULL);
            if (result != TCL_OK) goto done;
	    }
    }
  done:
    Tcl_DeleteCmdBuf(cmdBuf);
    return result;
}

int
Tcl_Startup(interp, argc, argv)
Tcl_Interp *interp;
int     argc;
char  **argv;
{
    int     c, i, result;
    char   *args;
    char   *interactive_session;

    /* merge the C main argv into a Tcl variable, argv, containing a list
     * of the argv elements */
    args = Tcl_Merge(argc, &argv[0]);
    Tcl_SetVar(interp, "argv", args, 1);
    ckfree(args);

    /* load the tcl startup code, this should pull in all of the tcl
     * procs, paths, command line processing, autoloads, packages, etc */
    result = Eval_Tcl_Startup(interp);
    if (result != TCL_OK) {
        print_result(stderr, result, interp->result);
        print_result(stderr, result, Tcl_GetVar(interp, "errorInfo", 1));
	fputs("Tcl initialzation failed!\n", stderr);
        exit(5);
    }

    /* if the invoked tcl interactively, as indicated by the tcl variable
     * interactiveSession not being a zero (should have been set by
     * the outcome of the startup call above, probably the loading of
     * TclInit.tcl, give the user an interactive session */
    interactive_session = Tcl_GetVar(interp, "interactiveSession", 1);
    if ((interactive_session != NULL) && 
        (strcmp(interactive_session, "0") != 0))
       Tcl_CommandLoop (interp, stdin, stdout, interactive_session);
}

