#
# Tcl master initialization file
#
# This file is loaded by all invocations of tcl.  
#
# it exits 0 on success, 1 for a non-Tcl error (Unix, etc),  2 for
#   source file not found, 3 for an error in Tcl source error
#
#

proc empty {list} {
    strequ $list ""
}

#
# autoload, provide support for demand-loaded procedures
#
proc autoload {file args} {
    global autoload_array
    foreach i $args {
	if ![info exists autoload_array($i)] {
	    set autoload_array($i) [list file $file]
	} else {
	    puts stderr "Warning: duplicate function \"$i\" between autoload file $file and $autoload_array($i)"
	}
    }
}

#
# package, defpackage - provide support for demand-loaded libraries
#
# load_package_indexes grinds through the Tcl pathlist and loads
#  all of the package's index files
#
proc package {args} {}
proc defpackage {packageName libName seekOffset Length args} {
    global packages autoload_array
    set packages($packageName) [list $libName $seekOffset $Length]
    foreach i $args {
	if ![info exists autoload_array($i)] {
	    set autoload_array($i) [list package $packageName]
	} else {
	    puts stderr "Warning: duplicate function \"$i\" between package $packageName file $libName and $autoload_array($i)"
	}
    }
}
proc load_package_indexes {} {
    global TCL_PATHLIST
    foreach dir $TCL_PATHLIST {
	if [strequ $dir "."] {set dir [pwd]}
	foreach pkg_file [globok $dir/*.tlib] {
	    set indexname [file $pkg_file root].tndx
	    if {![file $indexname exists] || ([file $indexname modify_time] < [file $pkg_file modify_time])} {
		    build_package_index $pkg_file
	    }
	    if [file $indexname exists] {
		source $indexname
	    }
	}
    }
}

#
# define PATHLIST as the PATH environment variable exploded into a Tcl list
# define TCL_PATHLIST from the TCLPATH Tcl variable as above, it will have
# been set by the default file or env var.
#
set PATHLIST [split : [getenv PATH]]
if ![info exists TCLPATH] {set TCLPATH ""}
if [empty $TCLPATH] {
    puts stderr "Warning - TCLPATH wasn't set by Tcl default file or env-var,"
    puts stderr "  autoload files and packages will only be found in current directory"
    set TCLPATH "."
}
set TCL_PATHLIST [split : $TCLPATH]

#
# searchpath - search a path list for a filename and if found, return
#  the path to the file, including the filename
proc searchpath {pathlist name} {
    foreach dir $pathlist {
	if [empty $dir] {set dir .}
	if [file $dir/$name exists] {
		return $dir/$name
	}
    }
    return ""
}

#
# load - do the same thing as "source", except look along the TCL_PATHLIST
#  for the file
#
proc load {name {offset {}} {length {}}} {
    global TCL_PATHLIST
    set where [searchpath $TCL_PATHLIST $name]
    if [empty $where] {
	error "couldn't find $name in Tcl search path"
    }
    if [empty $offset] {
	source $where
    } else {
	echo source $where $offset $length
	source $where $offset $length
    }
}

#
# demand_load - load in autoload file or package if named procedure 
#   is known from an autoload or from a package index file.  If it is
#   not known and the package indexes have not been loaded, load them
#   and try again (only occurs once at most)
#
set packageIndexesLoaded 0
proc demand_load {name} {
    global autoload_array packages packageIndexesLoaded
    if {!$packageIndexesLoaded && ![info exists autoload_array($name)]} {
	load_package_indexes
	set packageIndexesLoaded 1
    }
    if [info exists autoload_array($name)] {
	set autoload_record $autoload_array($name)
	if [strequ [index $autoload_record 0] file] {
	    load [index $autoload_record 1]
	} else {
	    set pkg_info $packages([index $autoload_record 1])
	    source [index $pkg_info 0] [index $pkg_info 1] [index $pkg_info 2]
	    if [empty [info procs [index $pkg_info 0]]] {return 0}
	}
	return 1
    }
    return 0
}

#
# procedure to handle invalid command names, first try to demand
# load them (autoload or package).  If that doesn't work,
# if the package indexes haven't been loaded, load them and
# check try to demand load again.  If that doesn't work,
# if Tcl is not at top level, return (i.e. bail), else see if there's a 
# file in the PATH of the same name and if so, define a proc that'll 
# execute it
#
proc invalid_command_proc {name} {
    global PATHLIST interactiveSession
    if [demand_load $name] return
    if {!$interactiveSession || ([info level] != 1)} return
    set binpath [searchpath $PATHLIST $name]
    if [empty $binpath] return
    proc $name {args} [format "execute_shell_command %s %s \$args" $name $binpath]
}

set invalid_command_name_hook invalid_command_proc

#
# define some useful autoloads...
#
autoload help.tcl help
autoload packages.tcl packages autoprocs build_package_index build_package_indexes 


#
#  put any code you want all Tcl programs to include here...
#


#
# if invoked as "tcl", if there are no command-line arguments, give the
# user an interactive session, else see if the first argument is "-f" and
# if so, make sure there is at least one following argument, take them
# out of the argv and do a source on the filename following the "-f".
#
# if the first arg is not -f, remove it from the argv and evaluate it
#
# else if we weren't invoked as "tcl", search along the path within the
# PATH environment variable for name.tcl, sourcing it if found, otherwise
# printing an error
#
set interactiveSession 0
set programName [file [replace argv 0] tail]
if {[strequ $programName tcl] || [strequ $programName xtcl]} {
    if {[length $argv] == 0} { 
	set interactiveSession 1
    } else {
	if [strequ [index $argv 0] -f] {
	    if {[length $argv] < 2} {
		error "filename required after '-f'"
	    }
	    replace argv 0
	    set progpath [replace argv 0]
	    set command {source $progpath}
	} else {
	    set command [replace argv 0]
	}
	if [catch $command errortext] {
	    if [strequ [index $errorCode 0] TCL] {
		puts stderr $errorInfo
		exit 3
	    } else {
		puts stderr $errortext
		exit 1
	    }
	}
    }
} else {
    # we were invoked under another name, find name.tcl
    if {[catch {load $programName.tcl} errortext] == 0} {
	if [strequ [index $errorCode 0] TCL] {
	    puts stderr $errorInfo
	    exit 3
	} else {
	    puts stderr $errortext
	    exit 1
	}
    } else {
	puts stderr [format "%s.tcl not found along search path" $programName]
	exit 2
    }
}

if !$interactiveSession {exit 0}

#
# put code you want interactive Tcl shells to include here...
#

echo "Welcome to" [custom_info name] [info tclversion]

#
# Define the toplevel prompt
#

set toplevel_prompt_hook {
    global programName
    return "$programName>"
    }

#
# Define the downlevel prompt
#

set downlevel_prompt_hook {return "=>"}

#
# define an invalid-command-name hook that, if you're at top level,
# a proc is defined of the command name you attempted that calls
# the temp_shell_command proc, which tries to execute your command
# as a "system" command, then deletes itself
#
# the "set a " on the system just keeps the return code from printing
#
proc execute_shell_command {name binpath arglist} {
    set a [system [concat $binpath $arglist]]
}

#
# if the user has a .tclrc file in their home directory, source it
#
set HOME [getenv HOME]
if [file $HOME/.tclrc exists] {source $HOME/.tclrc}

