#-----------------------------------------------------------------------------
#                                gen_help.tcl
#-----------------------------------------------------------------------------
#
# Program to extract help files from TCL manual pages or TCL script files.
# The help directories are built as a hierarchical tree of subjects and help
# files.  
#
# For nroff man pages, the areas of text to extract are delimited with:
#
#     '@help: subjectdir/helpfile
#     '@endhelp
#
# start in column one. The text between these markers is extracted and stored
# in help/subjectdir/help.  The file must not exists, this is done to enforced 
# cleaning out the directories before help file generation is started, thus
# removing any stale files.  The extracted text is run through:
#
#     nroff -man|col -xb
#
# If there is other text to include in the helpfile, but not in the manual 
# page, the text, along with nroff formatting commands, may be included using:
#
#     '@:Other text to include in the help page.
#
# If some header text, such as nroff macros, need to be preappended to the
# text streem before it is run through nroff, then that text can be bracketed
# with:
#
#     '@header
#     '@endheader
#
# If multiple header blocks are encountered, they will all be preappended.
#
# For TCL script files, which are indentified because they end in ".tcl",
# the text to be extracted is delimited by:
#
#    #@help: subjectdir/helpfile
#    #@endhelp
#
# The only processing done on text extracted from .tcl files it to replace
# the # in column one with a space.
#
#-----------------------------------------------------------------------------
# 
# To run this program:
#
#   tcl -f gen_help.tcl helpdir file-1 file-2 ...
#
# Where helpdir is the help tree root directory and file-n are the nroff
# manual pages or .tcl files to extract the help files from.  Helpdir should
# exists, but any subdirectories that don't exists will be created.  Helpdir
# should be cleaned up before the start of manual page generation, as this 
# program will not overwrite existing files.
#   
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Proc to ensure that all directories for the specified file path exists,
# and if they don't create them.

proc EnsureDirs {FilePath} {
    # Split the path into a list of files. If the path starts with /, then
    # the first element of the list is {}.  Nuke actual file name element.

    mkpath [file $FilePath dirname]
}


#-----------------------------------------------------------------------------
#
# Proc to extract nroff text to use as a header to all pass to nroff when
# processing a help file.
#    ManPageFH - The file handle of the manual page.
#

proc ExtractNroffHeader {ManPageFH} {
    global NroffHeader
    while {[gets $ManPageFH ManLine]} {
        if {[regexp ^'@endheader $ManLine]} {
            break;
            }
        if {[regexp ^'@: $ManLine]} {
            set ManLine [substr ManLine 3 end]
            }
        strcat NroffHeader "$ManLine\n"
        }
}

#-----------------------------------------------------------------------------
#
# Proc to extract a nroff help file when it is located in the text.
#    ManPageFH - The file handle of the manual page.
#    ManLine - The '@help: line starting the data to extract.
#

proc ExtractNroffHelp {ManPageFH ManLine} {
    global HelpDir NroffHeader
    set HelpFile "$HelpDir/[trim_string [substr $ManLine 7 end]]"
    if {[file $HelpFile exists]} {
        error "Help file already exists: $HelpFile"}
    EnsureDirs $HelpFile
    set HelpFH [open "| nroff -man|col -bx >$HelpFile" w]
    puts $HelpFH ".ad b"
    puts $HelpFH $NroffHeader
    while {[gets $ManPageFH ManLine]} {
        if {[regexp ^'@endhelp $ManLine]} {
            break;
            }
        if {[regexp ^'@: $ManLine]} {
            set ManLine [substr ManLine 3 end]
            }
        puts $HelpFH $ManLine
        }
    close $HelpFH
    chmod -w,+r $HelpFile
}

#-----------------------------------------------------------------------------
#
# Proc to extract a tcl script help file when it is located in the text.
#    ScriptPageFH - The file handle of the .tcl file.
#    ScriptLine - The #@help: line starting the data to extract.
#

proc ExtractScriptHelp {ScriptPageFH ScriptLine} {
    global HelpDir
    set HelpFile "$HelpDir/[trim_string [substr $ScriptLine 7 end]]"
    if {[file $HelpFile exists]} {
        error "Help file already exists: $HelpFile"}
    EnsureDirs $HelpFile
    set HelpFH [open $HelpFile w]
    while {[gets $ScriptPageFH ScriptLine]} {
        if {[regexp ^#@endhelp $ScriptLine]} {
            break;
            }
        if {[clength $ScriptLine] > 1} {
            set ScriptLine " [substr $ScriptLine 1 end]"
        } else {
            set ScriptLine ""
        }
        puts $HelpFH $ScriptLine
        }
    close $HelpFH
    chmod -w,+r $HelpFile
}

#-----------------------------------------------------------------------------
#
# Proc to scan a manual or .tcl file looking for the start of a help text
# sections and extracting those sections.
#    ManFileName - Name of file to extract.
#

proc ProcessManFile {ManFileName} {
   global NroffScanCT ScriptScanCT NroffHeader

   set ManPageFH [open $ManFileName r]
   if {[strequ [file $ManFileName extension] ".tcl"] || [strequ [file $ManFileName extension] ".tlib"]} {
       scan_file $ManPageFH $ScriptScanCT
   } else {
       set NroffHeader {}
       scan_file $ManPageFH $NroffScanCT
   }
   close $ManPageFH
}

#-----------------------------------------------------------------------------
# Main program body.

global HelpDir

if {[length $argv] < 2} {
    echo {Required arguments: helpdir manfile1 [manfile2..]}
    exit 1
    }

set HelpDir [replace argv 0]

if {![file $HelpDir isdirectory]} {
    echo "$HelpDir is not a directory or does not exist."  
    echo "This should be the help root directory"
    exit 1
    }

global NroffScanCT
set NroffScanCT [create_scancontext]
match "^'@help:" {
    global matchInfo
    ExtractNroffHelp $matchInfo(handle) $matchInfo(line)
    continue
    }

match "^'@header" {
    global matchInfo
    ExtractNroffHeader $matchInfo(handle)
    continue
    }

global ScriptScanCT
set ScriptScanCT [create_scancontext]
match "^#@help:" {
    global matchInfo
    ExtractScriptHelp $matchInfo(handle) $matchInfo(line)
    }


foreach ManFile $argv {
    if {[catch {ProcessManFile $ManFile} msg] != 0} {
        echo "Error processing manual file: $msg"
        echo $msg
        if {!$interactiveSession} {
            echo $errorInfo
            exit 1
            }
        }
    }
